#ifndef PARTICLE3D_INL
#define PARTICLE3D_INL

#include "Particle3D.h"
#include "LogSystem.h"
#include "Particles3DParameters.h"

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::Particle3DPool()
{
   LogSystemOut("Particle3D Pool Initialization...");
   
   memset(&m_Pool, 0, sizeof(m_Pool));
   memset(m_Pool.m_arrayParticlesSize, -1, sizeof(UINT) * NB_MAX_TOTAL_Particles);
   m_Pool.m_arrayParticlesSize[0] = NB_MAX_TOTAL_Particles;

   Particle3D* pParticles = m_Pool.m_arrayParticlesPool;
   Element3D* pElements = m_Pool.m_arrayElementsPool;
   for (UINT i = NB_MAX_TOTAL_Particles; i != 0; --i)
   {
      pParticles->m_pElement = pElements;
      pParticles++;
      pElements++;
   }

   LogSystemOut("Particle3D Pool Initialized!");
}
   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::~Particle3DPool()
{
   LogSystemOut("Particle3D Pool Destruction...");
   LogSystemOut("Particle3D Pool Destroyed!");
}

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  Particle3D* Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::AllocParticle3DPool(UINT _iNbParticles)
{
   LOGASSERT(_iNbParticles != 0, "AllocParticle3DPool : Can't allocate 0 Particles");
   
   Particle3D* pBest = NULL;
   UINT iBestSize = 0;
   UINT iBestIdx = 0;
   UINT idx = 0;
   while (idx < NB_MAX_TOTAL_Particles)
   {
      UINT iCurZoneSize = m_Pool.m_arrayParticlesSize[idx];
      if (m_Pool.m_arrayParticlesUse[idx] == false)
      {
         bool bIsOk = true;
         if (pBest != NULL)
         {
            if (_iNbParticles > iCurZoneSize)
            {
               bIsOk = false;
            }
            else if (iBestSize <= iCurZoneSize)
            {
               bIsOk = false;
            }
         }
         if (bIsOk)
         {
            pBest = m_Pool.m_arrayParticlesPool + idx;
            iBestSize = iCurZoneSize;
            iBestIdx = idx;
         }
      }
      idx += iCurZoneSize;
   }
   LOGASSERT(pBest != NULL, "AllocParticle3DPool : Can't Allocate Particles from Pool");
   
   m_Pool.m_arrayParticlesUse[iBestIdx] = true;
   m_Pool.m_arrayParticlesSize[iBestIdx] = _iNbParticles;
   if (_iNbParticles < iBestSize)
   {
      UINT iSubIdx = iBestIdx + _iNbParticles;
      m_Pool.m_arrayParticlesSize[iSubIdx] = iBestSize - _iNbParticles;      
      LOGASSERT(m_Pool.m_arrayParticlesUse[iSubIdx] == false, "AllocParticle3DPool : Particles Pool corrupted, use flag not false");
   }
   
   return pBest;   
}

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void        Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::FreeParticle3DPool(Particle3D* _pParticle3D)
{
   LOGASSERT(_pParticle3D != NULL, "FreeParticle3DPool : not a valid address (NULL)");
   LOGASSERT(m_Pool.m_arrayParticlesPool <= _pParticle3D, "FreeParticle3DPool : Address corrupted, not a ArrayParticlesPool valid address (_pParticle3D < m_Pool.m_arrayParticlesPool)");
   UINT iIdx = _pParticle3D - m_Pool.m_arrayParticlesPool;
   LOGASSERT(iIdx < NB_MAX_TOTAL_Particles, "FreeParticle3DPool : Address corrupted, not a ArrayParticlesPool valid address (iIdx >= Particles_POOL_MAX_Elements)");
   LOGASSERT(m_Pool.m_arrayParticlesUse[iIdx] == true, "FreeParticle3DPool : Particles Pool corrupted, use flag not true");
   m_Pool.m_arrayParticlesUse[iIdx] = false;
   UINT iCurZoneSize = m_Pool.m_arrayParticlesSize[iIdx];
   UINT iNextIdx = iIdx + iCurZoneSize;
   if ((iNextIdx < NB_MAX_TOTAL_Particles) && (m_Pool.m_arrayParticlesUse[iNextIdx] == false))
   {
      m_Pool.m_arrayParticlesSize[iIdx] += m_Pool.m_arrayParticlesSize[iNextIdx];
   }   
}

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  Emiter3D*     Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::AllocEmiter3DPool(UINT _iNbParticles)
{
   LOGASSERT(_iNbParticles != 0, "AllocEmiter3DPool : can't allocate 0 Particles Emiter3D");

   Emiter3D * pEmiter3D = m_Pool.m_arrayEmiter3DsPool;
   bool * pEmiter3DUse = m_Pool.m_arrayEmiter3DsUse;

   for (UINT i = NB_MAX_Emiter3DS; i != 0; --i)
   {
      if (!*pEmiter3DUse)
      {
         *pEmiter3DUse = true;
         pEmiter3D->m_iNbParticles = _iNbParticles;
         pEmiter3D->m_arrayParticles = AllocParticle3DPool(_iNbParticles);
         pEmiter3D->m_arrayElements = m_Pool.m_arrayElementsPool + (pEmiter3D->m_arrayParticles - m_Pool.m_arrayParticlesPool);
         
         pEmiter3D->m_pUserData = NULL;
         pEmiter3D->m_fctUpdateParticles = NULL;
         pEmiter3D->m_fctUpdateEmiter = NULL;
         return pEmiter3D;
      }
      pEmiter3D++;
      pEmiter3DUse++;
   }
   LOGASSERT(false, "AllocEmiter3DPool :  Can't Allocate Emiter3D from Pool");
   return NULL;
}

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void        Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::FreeEmiter3DPool(Emiter3D * _pEmiter3D)
{
   LOGASSERT(_pEmiter3D != NULL, "FreeEmiter3DPool : not a valid address (NULL)");

   int idx = _pEmiter3D - m_Pool.m_arrayEmiter3DsPool;
   m_Pool.m_arrayEmiter3DsUse[idx] = false;
   
   FreeParticle3DPool(_pEmiter3D->m_arrayParticles);
   
   _pEmiter3D->m_arrayParticles = NULL;
   _pEmiter3D->m_arrayElements = NULL;
   _pEmiter3D->m_pUserData = NULL;
   _pEmiter3D->m_fctUpdateParticles = NULL;
   _pEmiter3D->m_fctUpdateEmiter = NULL;
   _pEmiter3D->m_iNbParticles = 0;
}

   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  Emiter3D*  Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::AddEmiter(const vec4x& _vSourcePos, UINT _iNbParticles)
{
   Emiter3D * pEmiter3D = AllocEmiter3DPool(_iNbParticles);
   pEmiter3D->m_vSource3DPos = _vSourcePos;
   return pEmiter3D;
}
   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void     Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::RemoveEmiter(Emiter3D* _pEmiter3D)
{
   LOGASSERT(_pEmiter3D != NULL, "RemoveEmiter3D : not a valid address (NULL)");
   FreeEmiter3DPool(_pEmiter3D);
}

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline void Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::AddEmiter3DSorted(Emiter3D * pEmiter3D)
{
   Emiter3D ** pEmiter3DSorted = m_Pool.m_arrayEmiter3DsSorted;
   UINT nbEmiter3Ds = m_Pool.m_nbEmiter3DsToDisplay;
   UINT i = 0;
   for (; i < nbEmiter3Ds; ++i)
   {
      Emiter3D * pCurEmiter3D = *pEmiter3DSorted;
      if (pCurEmiter3D->m_fFar < pEmiter3D->m_fFar)
      {
         break;
      }      
      pEmiter3DSorted++;
   }
   if (i < nbEmiter3Ds)
   {
      Emiter3D ** pEmiter3DSorted2 = m_Pool.m_arrayEmiter3DsSorted + (nbEmiter3Ds - 1);
      for (int j = i; j < nbEmiter3Ds; j++)
      {
         *(pEmiter3DSorted2 + 1) = *pEmiter3DSorted2;
         pEmiter3DSorted2--;
      }
   }
   *pEmiter3DSorted = pEmiter3D;
   m_Pool.m_nbEmiter3DsToDisplay++;   
}

   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void     Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::DoAllWithoutDrawing(const Camera3D& _pCamera3D, UINT _iCurTime, const fixed_t& _fElapsedTime)
{
   UpdateEmiter3DsPool(_pCamera3D, _iCurTime, _fElapsedTime);
   UpdateParticlesPool(_pCamera3D, _iCurTime, _fElapsedTime);
}
   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void     Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::DrawPool(USHORT* _pData)
{
   Emiter3D ** ppEmiter3D = m_Pool.m_arrayEmiter3DsSorted;
   for (UINT i = m_Pool.m_nbEmiter3DsToDisplay; i != 0; --i)
   {
      Emiter3D * pEmiter3D = *ppEmiter3D;
      {
         LOGASSERT(m_Pool.m_arrayEmiter3DsUse[pEmiter3D - m_Pool.m_arrayEmiter3DsPool], "UpdatePool : Culling Error, Emiter3D not used => Not visible");
         DrawElement(pEmiter3D->m_arrayElements, pEmiter3D->m_iNbParticles, _pData);
      }
      ppEmiter3D++;
   }
}
   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void     Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::LogPool()
{
   // LOG OF ALL THE POOL
   char buffer[256];
   LogSystemOut("< Particles System LOG START >");

   {
      // Particles POOL LOGGING
      LogSystemOut("< Particles Allocation LOG >");
      {
         UINT iIdx = 0;
         while (iIdx < NB_MAX_TOTAL_Particles)
         {
            UINT iCurZoneSize = m_Pool.m_arrayParticlesSize[iIdx];
            bool bUsed = m_Pool.m_arrayParticlesUse[iIdx];
            
            sprintf(buffer, "Bloc : %d, Size : %d, Used : %d", iIdx, iCurZoneSize, bUsed);
            UINT iNxtIdx = iIdx + iCurZoneSize;
            for (UINT iCurIdx = iIdx + 1; iCurIdx < iNxtIdx; iCurIdx++)
            {
               if (m_Pool.m_arrayParticlesUse[iCurIdx] == true)
               {
                  sprintf(buffer, "Particles Allocation Error : Bloc corrupted at index %d", iCurIdx);               
                  LogSystemOut(buffer);
               }
            }
            iIdx = iNxtIdx;
         }
      }
   }
   LogSystemOut("< Particles System LOG END >"); 
}
   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void     Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::UpdateEmiter3DsPool(const Camera3D& _pCamera3D, UINT _iCurTime, const fixed_t& _fElapsedTime)
{
   Emiter3D * pEmiter3D = m_Pool.m_arrayEmiter3DsPool;
   bool * pEmiter3DUse = m_Pool.m_arrayEmiter3DsUse;

   m_Pool.m_nbEmiter3DsToDisplay = 0;
   for (UINT i = NB_MAX_Emiter3DS; i != 0; --i)
   {
      pEmiter3D->m_visible3D = false;
      if (*pEmiter3DUse)
      {
         if (pEmiter3D->m_fctUpdateEmiter)
            pEmiter3D->m_fctUpdateEmiter(pEmiter3D, _iCurTime, _fElapsedTime);
         DoEmiter3DCulling(pEmiter3D, _pCamera3D);
         if (pEmiter3D->m_visible3D)
         {
            AddEmiter3DSorted(pEmiter3D);
         }
      }
      pEmiter3D++;
      pEmiter3DUse++;
   }   
}
   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline  void     Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::UpdateParticlesPool(const Camera3D& _pCamera3D, UINT _iCurTime, const fixed_t& _fElapsedTime)
{
   Emiter3D ** ppEmiter3D = m_Pool.m_arrayEmiter3DsSorted;
   for (UINT i = m_Pool.m_nbEmiter3DsToDisplay; i != 0; --i)
   {
      Emiter3D * pEmiter3D = *ppEmiter3D;
      LOGASSERT(pEmiter3D, "UpdatePool : pEmiter3D == NULL");
      {
         LOGASSERT(m_Pool.m_arrayEmiter3DsUse[pEmiter3D - m_Pool.m_arrayEmiter3DsPool], "UpdatePool : Culling Error, Emiter3D not used => Not visible");
         LOGASSERT(pEmiter3D->m_fctUpdateParticles, "UpdatePool : Culling Error, Emiter3D don't have Update Function => Not visible");
         Particle3D * pParticle3D = pEmiter3D->m_arrayParticles;
         UINT iNbParticle3D = pEmiter3D->m_iNbParticles;
         for (UINT j = iNbParticle3D; j != 0; --j)
         {
            pEmiter3D->m_fctUpdateParticles(pEmiter3D, pParticle3D, _iCurTime, _fElapsedTime);            
            pParticle3D++;
         }
         Element3D3Dto2D(pEmiter3D->m_arrayElements, pEmiter3D->m_iNbParticles, _pCamera3D);
      }
      ppEmiter3D++;
   }   
}

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline void       Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::DoEmiter3DCulling(Emiter3D * _pEmiter3D, const Camera3D & _pCamera3D)
{
   LOGASSERT(_pEmiter3D != NULL, "DoEmiter3DCulling : _pEmiter3D == NULL");

   _pEmiter3D->m_visible3D = false;
   vec4x v2DPos = _pCamera3D.m_matViewProj * /*vec4x(*/_pEmiter3D->m_vSource3DPos/*, FIXED(1.0f))*/;  
   if ((UINT)v2DPos.w.intValue <= (EMITER3D_CULL_FAR << FIXED_DECAL))
   {
      fixed_t w = v2DPos.w;
      v2DPos.x /= w;
      v2DPos.y /= w;
      fixed_t f2DSize = _pEmiter3D->m_fSourceSize / w;

      if ((length<fixed_t>(vec2x(v2DPos.x, v2DPos.y)) - f2DSize).intValue <= (1 << FIXED_DECAL))
      {
         _pEmiter3D->m_fFar = w - f2DSize;
         _pEmiter3D->m_visible3D = true;
      }
   }
}

template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline void       Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::Element3D3Dto2D(Element3D* _arrayElements, UINT _iNbElements, const Camera3D& _pCamera3D)
{
       LOGASSERT(_arrayElements != NULL, "Element3D3Dto2D : _arrayElements == NULL");
       LOGASSERT(_iNbElements != 0, "Element3D3Dto2D : _iNbElements == 0");
       const mat4x& mVP = _pCamera3D.m_matViewProj;
      Element3D* elm = _arrayElements;
       for (UINT i = _iNbElements; i != 0; --i)
       {
           elm->m_2DPos = mVP * elm->m_3DPos;  
           elm->m_visible2D = false;
           if (elm->m_2DPos.w.intValue > 0)
           {
               elm->m_2DPos.x /= elm->m_2DPos.w;
               elm->m_2DPos.y /= elm->m_2DPos.w;
               
               fixed_t x = elm->m_2DPos.x+1;
               fixed_t y = elm->m_2DPos.y+1;
               if ((((UINT)x.intValue) <= ((UINT)2 << FIXED_DECAL)) && 
                     (((UINT)y.intValue) <= ((UINT)2 << FIXED_DECAL)))               
               {
                  UINT div = elm->m_2DPos.w.intValue / ELEMENT3D_LUMINOSITY_FAR;                   
                  int r, g, b;
                  r = (elm->m_3DColor >>  0) & 0x1F;
                  r -= div;
                  r = (r >= 0 ? r : 0); 
                  g = (elm->m_3DColor >>  6) & 0x1F;
                  g -= div;
                  g = (g >= 0 ? g : 0); 
                  b = (elm->m_3DColor >>  11) & 0x1F;
                  b -= div;
                  b = (b >= 0 ? b : 0);
                  elm->m_2DColor = (r << 0) | (g << 6) | (b << 11);
#ifdef Element3D_REMOVE_BLACK
                  if (elm->m_2DColor)
#endif
                     elm->m_visible2D = true;
               }
           }
           elm++;
       }   
}
   
template<UINT NB_MAX_Emiter3DS, UINT NB_MAX_TOTAL_Particles, Draw3DParticleFct FCT_DRAWParticle3D> inline void       Particle3DPool<NB_MAX_Emiter3DS, NB_MAX_TOTAL_Particles, FCT_DRAWParticle3D>::DrawElement(Element3D* _arrayElements, UINT _iNbElements, USHORT* _pData)
{
    LOGASSERT(_arrayElements != 0, "DrawElement3D : _arrayElements == NULL");
    LOGASSERT(_iNbElements != 0, "DrawElement3D : _iNbElements == 0");
    LOGASSERT(_pData != 0, "DrawElement3D : _pData == NULL");
    LOGASSERT(_iWidth != 0, "DrawElement3D : _iWidth == 0");
    LOGASSERT(_iHeight != 0, "DrawElement3D : _iHeight == 0");

    fixed_t fWidthDiv2 =  FIXED(((int)SCREEN_WIDTH / 2) - (ELEMENT3D_DRAW_MARGE * 2));
    fixed_t fHeightDiv2 = FIXED(((int)SCREEN_HEIGHT / 2) - (ELEMENT3D_DRAW_MARGE * 2));
    for (UINT i = _iNbElements; i != 0; --i)
    {
        Element3D& elm = *_arrayElements;
        _arrayElements++;
        if (elm.m_visible2D)
        {
            fixed_t x = elm.m_2DPos.x + FIXED(1);
            fixed_t y = elm.m_2DPos.y + FIXED(1);
            LOGASSERT((x >= FIXED(0)) && (x <= FIXED(2)) && (y >= FIXED(0)) && (y <= FIXED(2)), "DrawElement3D, Projection Error => point Out of the Screen but set visible");
            UINT xx = (UINT) ((ELEMENT3D_DRAW_MARGE + (x * fWidthDiv2)).intValue >> FIXED_DECAL);
            UINT yy = (UINT) ((ELEMENT3D_DRAW_MARGE + (y * fHeightDiv2)).intValue >> FIXED_DECAL);
            UINT idx = xx + (yy * SCREEN_WIDTH);  
            FCT_DRAWParticle3D(xx + (yy * SCREEN_WIDTH), elm.m_2DPos.z, elm.m_2DColor, _pData);          
        }
    }       
}

#endif
