#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <SDL.h>

#include "utils.h"
#include "utils2D.h"
#include "utils3D.h"
#include "Particle2D.h"
#include "Particles2DDrawSamples.h"
#include "Particle3D.h"
#include "Particles3DDrawSamples.h"
#include "LogSystem.h"
#include "FpsCounter.h"

// PARTICULES FUNCTIONS
#include "Fontaine.h"
#include "Feu.h"
#include "Comete.h"

#include "Fontaine3D.h"
#include "Feu3D.h"
#include "Comete3D.h"

#ifdef GP2X
#include "ramhack.h"
#endif

/* GP2X button mapping */
enum MAP_KEY
{
	VK_UP         , // 0
	VK_UP_LEFT    , // 1
	VK_LEFT       , // 2
	VK_DOWN_LEFT  , // 3
	VK_DOWN       , // 4
	VK_DOWN_RIGHT , // 5
	VK_RIGHT      , // 6
	VK_UP_RIGHT   , // 7
	VK_START      , // 8
	VK_SELECT     , // 9
	VK_FL         , // 10
	VK_FR         , // 11
	VK_FA         , // 12
	VK_FB         , // 13
	VK_FX         , // 14
	VK_FY         , // 15
	VK_VOL_UP     , // 16
	VK_VOL_DOWN   , // 17
	VK_TAT          // 18
};

/* The screen surface, joystick device */
SDL_Surface *screen = NULL;
SDL_Joystick *joy = NULL;

#define  POOL_TYPE  ParticlePool<50, 3000, DrawParticles_SET_BLUR8>    
#define  POOL3D_TYPE  Particle3DPool<50, 3000, DrawParticles3D_SET_BLUR8>    

POOL_TYPE *g_2DPool = NULL;
POOL3D_TYPE *g_3DPool = NULL;

void Terminate(void)
{
	LogSystemOut("Terminate !");
   
   if (g_2DPool)
   {
      g_2DPool->LogPool();
      delete g_2DPool;
   }
   
   if (g_3DPool)
   {
      g_3DPool->LogPool();
      delete g_3DPool;
   }
   
   LogSystemEnd();

	SDL_Quit();
#ifdef GP2X
   mmuunhack();
   mmsp2_regs_deinit();

	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
#endif
}

int main (int argc, char *argv[])
{
	int done;
	/* Initialize Log System */
	LogSystemStart("ParticlesLog.txt");
	
	/* Initialize Particles Pool */
   g_2DPool = new POOL_TYPE();
	g_3DPool = new POOL3D_TYPE();

	/* Initialize SDL */
	if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK) < 0) {
		fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
		exit (1);
	}
	atexit (Terminate);

	SDL_ShowCursor(SDL_DISABLE);

	/* Set 320x240 16-bits video mode */
	screen = SDL_SetVideoMode (320, 240, 16, SDL_SWSURFACE);
	if (screen == NULL) {
		fprintf (stderr, "Couldn't set 320x240x16 video mode: %s\n", SDL_GetError ());
		exit (2);
	}

	/* Check and open joystick device */
	if (SDL_NumJoysticks() > 0) {
		joy = SDL_JoystickOpen(0);
		if(!joy) {
			fprintf (stderr, "Couldn't open joystick 0: %s\n", SDL_GetError ());
		}
	}

#ifdef GP2X
	/* Only use GP2X code here */
	mmuhack();
	mmsp2_regs_init();
//	ramhack();

//   GP2XSetGamma(120);
#endif

#ifdef WIN32
	/* Only use Windows code here */
#endif
// PREPARATION DES EMITERS
#define NB_EMITER       12
   Emiter* emiter2D[NB_EMITER];

// 5 x 2 FONTAINES
   for (int i = 0; i < 10; i += 2)
   {
      emiter2D[i] = g_2DPool->AddEmiter(vec2D(rand() % SCREEN_WIDTH, rand() % SCREEN_HEIGHT) * 10, 100);
      emiter2D[i]->m_fctUpdateParticles = FontaineParticles;
      emiter2D[i]->m_fctUpdateEmiter = FontaineEmiters;
      emiter2D[i]->m_fSourceSize = (300);
      FontaineParticlesInit(emiter2D[i]);

      emiter2D[i+1] = g_2DPool->AddEmiter(vec2D(rand() % SCREEN_WIDTH, rand() % SCREEN_HEIGHT) * 10, 100);
      emiter2D[i+1]->m_fctUpdateParticles = FontaineParticles;
      emiter2D[i+1]->m_fctUpdateEmiter = FontaineEmiters;
      emiter2D[i+1]->m_fSourceSize = (300);
      FontaineParticlesInit(emiter2D[i+1]);
   }
// 1 FEU
   emiter2D[10] = g_2DPool->AddEmiter(vec2D((SCREEN_WIDTH/2), (SCREEN_HEIGHT-1)) * 10, 1000);
   emiter2D[10]->m_fctUpdateParticles = FeuParticles;
   emiter2D[10]->m_fctUpdateEmiter = NULL;
   emiter2D[10]->m_fSourceSize = (500);
   FeuParticlesInit(emiter2D[10]);
// 1 COMETE
   emiter2D[11] = g_2DPool->AddEmiter(vec2D((SCREEN_WIDTH/2), (SCREEN_HEIGHT/2)) * 10, 500);
   emiter2D[11]->m_fctUpdateParticles = CometeParticles;
   emiter2D[11]->m_fctUpdateEmiter = CometeEmiters;
   emiter2D[11]->m_fSourceSize = (500);
   CometeParticlesInit(emiter2D[11]);


   Emiter3D* emiter[NB_EMITER];

// 5 x 2 FONTAINES
   for (int i = 0; i < 10; i += 2)
   {
      emiter[i] = g_3DPool->AddEmiter(vec4x(FIXED(-1), FIXED(i), FIXED(0), FIXED(1)), 150);
      emiter[i]->m_fctUpdateParticles = Fontaine3DParticles;
      emiter[i]->m_fctUpdateEmiter = NULL;
      emiter[i]->m_fSourceSize = FIXED(3);
      Fontaine3DParticlesInit(emiter[i]);

      emiter[i+1] = g_3DPool->AddEmiter(vec4x(FIXED(1), FIXED(i), FIXED(0), FIXED(1)), 150);
      emiter[i+1]->m_fctUpdateParticles = Fontaine3DParticles;
      emiter[i+1]->m_fctUpdateEmiter = NULL;
      emiter[i+1]->m_fSourceSize = FIXED(3);
      Fontaine3DParticlesInit(emiter[i+1]);
   }
// 1 FEU
   emiter[10] = g_3DPool->AddEmiter(vec4x(FIXED(0), FIXED(-2), FIXED(0), FIXED(1)), 1000);
   emiter[10]->m_fctUpdateParticles = Feu3DParticles;
   emiter[10]->m_fctUpdateEmiter = NULL;
   emiter[10]->m_fSourceSize = FIXED(5);
   Feu3DParticlesInit(emiter[10]);
// 1 COMETE
   emiter[11] = g_3DPool->AddEmiter(vec4x(FIXED(0), FIXED(-4), FIXED(2), FIXED(1)), 500);
   emiter[11]->m_fctUpdateParticles = Comete3DParticles;
   emiter[11]->m_fctUpdateEmiter = Comete3DEmiters;
   emiter[11]->m_fSourceSize = FIXED(5);
   Comete3DParticlesInit(emiter[11]);
   
   Camera camera2D;
   camera2D.m_3DPos = vec2D((0), (0));
   camera2D.m_iDiv = 10 * PARTICLES_DIVZ;


   Camera3D camera;
//   camera.m_matView = lookat_matrix<fixed_t>(vec4x(FIXED(-1), FIXED(-1), FIXED(1)), vec4x(FIXED(1), FIXED(1), FIXED(1)), vec4x(FIXED(0), FIXED(0), FIXED(1)));
   camera.m_matProj = perspective_matrix<fixed_t>(FIXED(60), FIXED(4.0f/3.0f), FIXED(0.1f), FIXED(100.0f));
//   camera.m_matViewProj = camera.m_matProj * camera.m_matView;
   camera.m_3DPos = vec3x(FIXED(0), FIXED(0), FIXED(0));

   int iLastTime = SDL_GetTicks();
   int iCurTime = iLastTime;
   int iElapsedTime;
   
#define CAM_TIME_TO_GO_NEXT   4000
	vec2D vDestCamPos2D = vec2D((10), (10));
   vec2D vLastCamPos2D = vDestCamPos2D;

	vec3x vDestCamPos = vec3x(FIXED(0.125f), FIXED(1), FIXED(0.5f));
   vDestCamPos = normalize<fixed_t>(vDestCamPos);
   vDestCamPos.x *= 13;
   vDestCamPos.y *= 13;
   vDestCamPos.z *= 13;
   vec3x vLastCamPos = vDestCamPos;
   int   iNxtCamChange = iCurTime;
   
   camera.m_3DPos = vDestCamPos;
   
   vec3x vLookAtPos = vec3x(FIXED(0), FIXED(0), FIXED(0));
   vec3x vUp = vec3x(FIXED(0), FIXED(0), FIXED(-1));
   vec3x vRight = vec3x(FIXED(1), FIXED(0), FIXED(0));

   char buffer[128];
   int flag = 1;
	done = 0;
	while (!done)
	{            
      /* Time Computing */
      iLastTime = iCurTime;
      iCurTime = SDL_GetTicks();
      iElapsedTime = iCurTime - iLastTime;
 
		SDL_Event event;
		
		/* Check for events */
		while (SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					/* if press Ctrl + C, terminate program */
					if ( (event.key.keysym.sym == SDLK_c) && (event.key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL)) )
					{
						done = 1;
						LogSystemOut("Quit game from KeyDown !");
               }
					if ( event.key.keysym.sym == SDLK_SPACE )
					{
                  flag = (flag % 3) + 1;
               }
               break;
				case SDL_KEYUP:
					break;
				case SDL_JOYBUTTONDOWN:
					/* if press Start button, terminate program */
					if ( event.jbutton.button == VK_START )
					{
						done = 1;
 						LogSystemOut("Quit game from Start Button !");
               }
					if ( event.jbutton.button == VK_SELECT )
					{
                  flag = (flag % 3) + 1;
               }
					break;
				case SDL_JOYBUTTONUP:
					break;
				case SDL_QUIT:
               {
   					done = 1;
 						LogSystemOut("Quit game from SDL_QUIT !");
               }
					break;
				default:
					break;
			}
		}
		/* Processing */	

#ifndef NOTFIXED_CAMERA
      if (iNxtCamChange <= iCurTime)
      {
         vLastCamPos = vDestCamPos;
         
         vDestCamPos = vec3x(FIXED(-64 + (rand() & 127)), FIXED(-64 + (rand() & 127)), FIXED(-64 + (rand() & 127)));
         vDestCamPos = normalize<fixed_t>(vDestCamPos);
         vDestCamPos.x *= FIXED(25);
         vDestCamPos.y *= FIXED(25);
         vDestCamPos.z *= FIXED(25);
         
         vLastCamPos2D = vDestCamPos2D;      
         vDestCamPos2D = vec2D(SCREEN_WIDTH * 10, SCREEN_HEIGHT * 10) - vDestCamPos2D;         
         
         iNxtCamChange = iCurTime + CAM_TIME_TO_GO_NEXT;
      }
      fixed_t coef1 = FIXED(iNxtCamChange - iCurTime) / FIXED(CAM_TIME_TO_GO_NEXT);
      fixed_t coef2 = FIXED(1) - coef1;
      UINT coef2D1 = (iNxtCamChange - iCurTime);
      UINT coef2D2 = (CAM_TIME_TO_GO_NEXT) - coef2D1;
#else  
      UINT coef2D1 = (CAM_TIME_TO_GO_NEXT);
      UINT coef2D2 = (0);
      fixed_t coef1 = FIXED(1);
      fixed_t coef2 = FIXED(0);
#endif

		camera2D.m_3DPos.x = ((vLastCamPos2D.x * coef2D1) + (vDestCamPos2D.x * coef2D2)) / CAM_TIME_TO_GO_NEXT;
		camera2D.m_3DPos.y = ((vLastCamPos2D.y * coef2D1) + (vDestCamPos2D.y * coef2D2)) / CAM_TIME_TO_GO_NEXT;
static int sensZ = 1;
      camera2D.m_iDiv += sensZ * iElapsedTime * (PARTICLES_DIVZ>>8);
      if (camera2D.m_iDiv >= 20 * PARTICLES_DIVZ)
         sensZ = -1;		
      if (camera2D.m_iDiv <= 5 * PARTICLES_DIVZ)
         sensZ = 1;		


		camera.m_3DPos.x = (vLastCamPos.x * coef1) + (vDestCamPos.x * coef2);
		camera.m_3DPos.y = (vLastCamPos.y * coef1) + (vDestCamPos.y * coef2);
		camera.m_3DPos.z = (vLastCamPos.z * coef1) + (vDestCamPos.z * coef2);
		
	   camera.m_matView = lookat_matrix<fixed_t>(camera.m_3DPos, vLookAtPos, vUp);
      camera.m_matViewProj = camera.m_matProj * camera.m_matView;

      if (flag & 1)
         g_2DPool->DoAllWithoutDrawing(camera2D, iCurTime, iElapsedTime);

      if (flag & 2)
         g_3DPool->DoAllWithoutDrawing(camera, iCurTime, FIXED(iElapsedTime) / FIXED(1000));
      		
		static int color = 0;
		SDL_FillRect(screen, NULL, COLOR_RGB16(color>>8, color>>8, color>>8)); color++;
		
		SDL_LockSurface(screen);
	
      if (flag & 1)
   		g_2DPool->DrawPool((USHORT*)screen->pixels);

      if (flag & 2)
   		g_3DPool->DrawPool((USHORT*)screen->pixels);

   	SDL_UnlockSurface(screen);

		SDL_Flip(screen);

      UpdateFPS(iElapsedTime);
	}

	LogSystemOut("Quitting...");

   for (int i = 0; i < NB_EMITER; i++)
   {
      SAFE_DELETE_ARRAY((vec2D*)emiter2D[i]->m_pUserData);
      g_2DPool->RemoveEmiter(emiter2D[i]);
      
      SAFE_DELETE_ARRAY((vec4x*)emiter[i]->m_pUserData);
      g_3DPool->RemoveEmiter(emiter[i]);
   }

	LogSystemOut("Quit !");

	return 0;
}
