#ifndef PARTICLE2D_INL
#define PARTICLE2D_INL

#include "Particle2D.h"
#include "LogSystem.h"
#include "Particles2DParameters.h"

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::ParticlePool()
{
   LogSystemOut("Particle Pool Initialization...");
   
   memset(&m_Pool, 0, sizeof(m_Pool));
   memset(m_Pool.m_arrayParticlesSize, -1, sizeof(UINT) * NB_MAX_TOTAL_PARTICLES);
   m_Pool.m_arrayParticlesSize[0] = NB_MAX_TOTAL_PARTICLES;
   
   Particle* pParticles = m_Pool.m_arrayParticlesPool;
   Element* pElements = m_Pool.m_arrayElementsPool;
   for (UINT i = NB_MAX_TOTAL_PARTICLES; i != 0; --i)
   {
      pParticles->m_pElement = pElements;
      pParticles++;
      pElements++;
   } 

   LogSystemOut("Particle Pool Initialized!");
}
   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::~ParticlePool()
{
   LogSystemOut("Particle Pool Destruction...");
   LogSystemOut("Particle Pool Destroyed!");
}

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  Particle* ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::AllocParticlePool(UINT _iNbParticles)
{
   LOGASSERT(_iNbParticles != 0, "AllocParticlePool : Can't allocate 0 particles");
   
   Particle* pBest = NULL;
   UINT iBestSize = 0;
   UINT iBestIdx = 0;
   UINT idx = 0;
   while (idx < NB_MAX_TOTAL_PARTICLES)
   {
      UINT iCurZoneSize = m_Pool.m_arrayParticlesSize[idx];
      if (m_Pool.m_arrayParticlesUse[idx] == false)
      {
         bool bIsOk = true;
         if (pBest != NULL)
         {
            if (_iNbParticles > iCurZoneSize)
            {
               bIsOk = false;
            }
            else if (iBestSize <= iCurZoneSize)
            {
               bIsOk = false;
            }
         }
         if (bIsOk)
         {
            pBest = m_Pool.m_arrayParticlesPool + idx;
            iBestSize = iCurZoneSize;
            iBestIdx = idx;
         }
      }
      idx += iCurZoneSize;
   }
   LOGASSERT(pBest != NULL, "AllocParticlePool : Can't Allocate Particles from Pool");
   
   m_Pool.m_arrayParticlesUse[iBestIdx] = true;
   m_Pool.m_arrayParticlesSize[iBestIdx] = _iNbParticles;
   if (_iNbParticles < iBestSize)
   {
      UINT iSubIdx = iBestIdx + _iNbParticles;
      m_Pool.m_arrayParticlesSize[iSubIdx] = iBestSize - _iNbParticles;      
      LOGASSERT(m_Pool.m_arrayParticlesUse[iSubIdx] == false, "AllocParticlePool : Particles Pool corrupted, use flag not false");
   }
   
   return pBest;
}

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void        ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::FreeParticlePool(Particle* _pParticle)
{
   LOGASSERT(_pParticle != NULL, "FreeParticlePool : not a valid address (NULL)");
   LOGASSERT(m_Pool.m_arrayParticlesPool <= _pParticle, "FreeParticlePool : Address corrupted, not a ArrayParticlesPool valid address (_pParticle < m_Pool.m_arrayParticlesPool)");
   UINT iIdx = _pParticle - m_Pool.m_arrayParticlesPool;
   LOGASSERT(iIdx < PARTICLES_POOL_MAX_ELEMENTS, "FreeParticlePool : Address corrupted, not a ArrayParticlesPool valid address (iIdx >= PARTICLES_POOL_MAX_ELEMENTS)");
   LOGASSERT(m_Pool.m_arrayParticlesUse[iIdx] == true, "FreeParticlePool : Particles Pool corrupted, use flag not true");
   m_Pool.m_arrayParticlesUse[iIdx] = false;
   UINT iCurZoneSize = m_Pool.m_arrayParticlesSize[iIdx];
   UINT iNextIdx = iIdx + iCurZoneSize;
   if ((iNextIdx < NB_MAX_TOTAL_PARTICLES) && (m_Pool.m_arrayParticlesUse[iNextIdx] == false))
   {
      m_Pool.m_arrayParticlesSize[iIdx] += m_Pool.m_arrayParticlesSize[iNextIdx];
   }
}

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  Emiter*     ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::AllocEmiterPool(UINT _iNbParticles)
{
   LOGASSERT(_iNbParticles != 0, "AllocEmiterPool : can't allocate 0 Particles Emiter");

   Emiter * pEmiter = m_Pool.m_arrayEmitersPool;
   bool * pEmiterUse = m_Pool.m_arrayEmitersUse;

   for (UINT i = NB_MAX_EMITERS; i != 0; --i)
   {
      if (!*pEmiterUse)
      {
         *pEmiterUse = true;
         pEmiter->m_iNbParticles = _iNbParticles;
         pEmiter->m_arrayParticles = AllocParticlePool(_iNbParticles);
         pEmiter->m_arrayElements = m_Pool.m_arrayElementsPool + (pEmiter->m_arrayParticles - m_Pool.m_arrayParticlesPool);
         
         pEmiter->m_pUserData = NULL;
         pEmiter->m_fctUpdateParticles = NULL;
         pEmiter->m_fctUpdateEmiter = NULL;
         return pEmiter;
      }
      pEmiter++;
      pEmiterUse++;
   }
   LOGASSERT(false, "AllocEmiterPool :  Can't Allocate Emiter from Pool");
   return NULL;
}

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void        ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::FreeEmiterPool(Emiter * _pEmiter)
{
   LOGASSERT(_pEmiter != NULL, "FreeEmiterPool : not a valid address (NULL)");

   int idx = _pEmiter - m_Pool.m_arrayEmitersPool;
   m_Pool.m_arrayEmitersUse[idx] = false;
   
   FreeParticlePool(_pEmiter->m_arrayParticles);
   
   _pEmiter->m_arrayParticles = NULL;
   _pEmiter->m_arrayElements = NULL;
   _pEmiter->m_pUserData = NULL;
   _pEmiter->m_fctUpdateParticles = NULL;
   _pEmiter->m_fctUpdateEmiter = NULL;
   _pEmiter->m_iNbParticles = 0;
}

   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  Emiter*  ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::AddEmiter(const vec2D& _vSourcePos, UINT _iNbParticles)
{
   Emiter * pEmiter = AllocEmiterPool(_iNbParticles);
   pEmiter->m_vSource3DPos = _vSourcePos;
   return pEmiter;
  
}
   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void     ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::RemoveEmiter(Emiter* _pEmiter)
{
   LOGASSERT(_pEmiter != NULL, "RemoveEmiter : not a valid address (NULL)");
   FreeEmiterPool(_pEmiter);
}

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline void ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::AddEmiterSorted(Emiter * pEmiter)
{
   m_Pool.m_arrayEmitersSorted[m_Pool.m_nbEmitersToDisplay++] = pEmiter;
}

   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void     ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::DoAllWithoutDrawing(const Camera& _pCamera, UINT _iCurTime, UINT _fElapsedTime)
{
   UpdateEmitersPool(_pCamera, _iCurTime, _fElapsedTime);
   UpdateParticlesPool(_pCamera, _iCurTime, _fElapsedTime);
}
   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void     ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::DrawPool(USHORT* _pData)
{
   Emiter ** ppEmiter = m_Pool.m_arrayEmitersSorted;
   for (UINT i = m_Pool.m_nbEmitersToDisplay; i != 0; --i)
   {
      Emiter * pEmiter = *ppEmiter;
      {
         LOGASSERT(m_Pool.m_arrayEmitersUse[pEmiter - m_Pool.m_arrayEmitersPool], "UpdatePool : Culling Error, Emiter not used => Not visible");
         DrawElement(pEmiter->m_arrayElements, pEmiter->m_iNbParticles, _pData);
      }
      ppEmiter++;
   }   
}
   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void     ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::LogPool()
{
   // LOG OF ALL THE POOL
   char buffer[256];
   LogSystemOut("< Particles System LOG START >");

   {
      // PARTICLES POOL LOGGING
      LogSystemOut("< Particles Allocation LOG >");
      {
         UINT iIdx = 0;
         while (iIdx < NB_MAX_TOTAL_PARTICLES)
         {
            UINT iCurZoneSize = m_Pool.m_arrayParticlesSize[iIdx];
            bool bUsed = m_Pool.m_arrayParticlesUse[iIdx];
            
            sprintf(buffer, "Bloc : %d, Size : %d, Used : %d", iIdx, iCurZoneSize, bUsed);
            UINT iNxtIdx = iIdx + iCurZoneSize;
            for (UINT iCurIdx = iIdx + 1; iCurIdx < iNxtIdx; iCurIdx++)
            {
               if (m_Pool.m_arrayParticlesUse[iCurIdx] == true)
               {
                  sprintf(buffer, "Particles Allocation Error : Bloc corrupted at index %d", iCurIdx);               
                  LogSystemOut(buffer);
               }
            }
            iIdx = iNxtIdx;
         }
      }
   }
   LogSystemOut("< Particles System LOG END >"); 
}
   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void     ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::UpdateEmitersPool(const Camera& _pCamera, UINT _iCurTime, UINT _fElapsedTime)
{
   Emiter * pEmiter = m_Pool.m_arrayEmitersPool;
   bool * pEmiterUse = m_Pool.m_arrayEmitersUse;

   m_Pool.m_nbEmitersToDisplay = 0;
   for (UINT i = NB_MAX_EMITERS; i != 0; --i)
   {
      pEmiter->m_visible3D = false;
      if (*pEmiterUse)
      {
         UpdateEmiterFct fct = pEmiter->m_fctUpdateEmiter;
         if (fct)
            fct(pEmiter, _iCurTime, _fElapsedTime);
         DoEmiterCulling(pEmiter, _pCamera);
         if (pEmiter->m_visible3D)
         {
            AddEmiterSorted(pEmiter);
         }
      }
      pEmiter++;
      pEmiterUse++;
   }
}
   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline  void     ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::UpdateParticlesPool(const Camera& _pCamera, UINT _iCurTime, UINT _fElapsedTime)
{
   Emiter ** ppEmiter = m_Pool.m_arrayEmitersSorted;
   for (UINT i = m_Pool.m_nbEmitersToDisplay; i != 0; --i)
   {
      Emiter * pEmiter = *ppEmiter;
      LOGASSERT(pEmiter, "UpdatePool : pEmiter == NULL");
      {
         LOGASSERT(m_Pool.m_arrayEmitersUse[pEmiter - m_Pool.m_arrayEmitersPool], "UpdatePool : Culling Error, Emiter not used => Not visible");
         LOGASSERT(pEmiter->m_fctUpdateParticles, "UpdatePool : Culling Error, Emiter don't have Update Function => Not visible");
         Particle * pParticle = pEmiter->m_arrayParticles;
         UINT iNbParticle = pEmiter->m_iNbParticles;
         UpdateParticleFct fct = pEmiter->m_fctUpdateParticles;
         for (UINT j = iNbParticle; j != 0; --j)
         {
            fct(pEmiter, pParticle, _iCurTime, _fElapsedTime);            
            pParticle++;
         }
         Element3Dto2D(pEmiter->m_arrayElements, iNbParticle, _pCamera);
      }
      ppEmiter++;
   }
}

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline void       ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::DoEmiterCulling(Emiter * _pEmiter, const Camera & _pCamera)
{
   LOGASSERT(_pEmiter != NULL, "DoEmiterCulling : _pEmiter == NULL");

   _pEmiter->m_visible3D = false;
   vec2D v2DPos = ((_pEmiter->m_vSource3DPos - _pCamera.m_3DPos) * PARTICLES_DIVZ) / _pCamera.m_iDiv;  
   v2DPos.x += (SCREEN_WIDTH / 2);
   v2DPos.y += (SCREEN_HEIGHT / 2);
   {
      UINT iSize = (_pEmiter->m_fSourceSize * PARTICLES_DIVZ) / _pCamera.m_iDiv;
      if ((((UINT)(v2DPos.x + iSize)) < (SCREEN_WIDTH + iSize + iSize)) && (((UINT)v2DPos.y + iSize) < (SCREEN_HEIGHT + iSize + iSize)))
      {
         _pEmiter->m_visible3D = true;
      }
   }   
}

template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline void       ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::Element3Dto2D(Element* _arrayElements, UINT _iNbElements, const Camera& _pCamera)
{
   LOGASSERT(_arrayElements != NULL, "Element3Dto2D : _arrayElements == NULL");
   LOGASSERT(_iNbElements != 0, "Element3Dto2D : _iNbElements == 0");
   Element* elm = _arrayElements;
   for (UINT i = _iNbElements; i != 0; --i)
   {
     vec2D vPos = ((elm->m_3DPos - _pCamera.m_3DPos) * PARTICLES_DIVZ) / _pCamera.m_iDiv;  
     vPos.x += (SCREEN_WIDTH / 2);
     vPos.y += (SCREEN_HEIGHT / 2);
     elm->m_2DPos = vPos;
     elm->m_visible2D = false;
     {
         if ((((UINT)(vPos.x-1)) < (SCREEN_WIDTH-2)) && (((UINT)(vPos.y-1)) < (SCREEN_HEIGHT-2)))
         {
            elm->m_visible2D = true;
         }
     }
     elm++;
   }   
}
   
template<UINT NB_MAX_EMITERS, UINT NB_MAX_TOTAL_PARTICLES, DrawParticleFct FCT_DRAWPARTICLE> inline void       ParticlePool<NB_MAX_EMITERS, NB_MAX_TOTAL_PARTICLES, FCT_DRAWPARTICLE>::DrawElement(Element* _arrayElements, UINT _iNbElements, USHORT* _pData)
{
    LOGASSERT(_arrayElements != 0, "DrawElement : _arrayElements == NULL");
    LOGASSERT(_iNbElements != 0, "DrawElement : _iNbElements == 0");
    LOGASSERT(_pData != 0, "DrawElement : _pData == NULL");

    for (UINT i = _iNbElements; i != 0; --i)
    {
        Element& elm = *_arrayElements;
        _arrayElements++;
        if (elm.m_visible2D)
        {
            FCT_DRAWPARTICLE(elm.m_2DPos.x + (elm.m_2DPos.y * SCREEN_WIDTH), elm.m_3DColor, _pData);            
        }
    }          
}

#endif
